/* 
 In the following threaded program using pthreads,
 the two threads try to add to the "log" variable
 at the same time.  On my machine, this program almost
 always crashes, because both threads cannot access 
 "log" simultaniously.  How could you change add_to_log
 to make this program always work properly?
 
 NOTE: 
   This is a POSIX pthreads program, so it won't build
   on Windows-- only UNIX machines.  You'll probably have
   to link this program with "-lpthread", like this:
     g++ problem1.cpp -lpthread
 
 Orion Lawlor, olawlor@acm.org, 2005/2/20 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <string>

/************** Logging Module ***************/
/* Global "log" variable: stores history of what happened. */
std::string log;
pthread_mutex_t log_lock=PTHREAD_MUTEX_INITIALIZER;

/* Add this string as an entry in the log. */
void log_entry(const char *newLogItem) {
	??
	log+=newLogItem;
	??
}


/**************** Threads *****************/
/* Number of iterations to run A and B for. */
int nIter=5;
volatile int A_done=0; /* silly way to tell when A has finished */

void *runA(void *p) {
	int i,n;
	for (n=0;n<nIter;n++) {
		for (i=0;i<1000*1000;i++) log_entry("A");
		printf("A (%d)\n",n);
	}
	A_done=1;
}

void *runB(void *p) {
	int i,n;
	for (n=0;n<nIter;n++) {
		for (i=0;i<1000*1000;i++) log_entry("B");
		printf("      B (%d)\n",n);
	}
}

int main() {
	pthread_t A;
	pthread_create(&A,0,runA,0);
	runB(0);
	while (!A_done) { /* silly-wait until thread A is done */ }
	
	printf("Finished: log has %d items\n",log.size());
	if (log.size()==2*nIter*1000*1000)
		printf("The log is the right size!  Problem complete!\n");
	return 0;
}
