/* 
 In the following user-level thread program, we'd like
 the thread "A" and the thread "B" to alternate execution.
 What single line of code could we add at the ?? lines
 in runA and runB to switch to the other thread?
 
 NOTE: 
   This program uses the POSIX 1003.1-2001 user-level
   thread routines.  It can be compiled on any modern 
   UNIX machine like this:
     gcc problem1.c
 
 Orion Lawlor, olawlor@acm.org, 2005/2/20 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>
#include <ucontext.h>

/* Tiny little user-level thread library */
typedef void (*threadFn)(void *);
typedef void (*void_fn)(void);

/* Create a new user-level thread to run this function */
void makeThread(ucontext_t *T,threadFn fn,void *ptr) {
	getcontext(T);
	/* Allocate a stack for the new thread */
	int stackLen=32*1024;
	T->uc_stack.ss_sp=malloc(stackLen);
	T->uc_stack.ss_size=stackLen;
	T->uc_stack.ss_flags=0;
	makecontext(T,(void_fn)fn,1,ptr); /* sets up a user-level thread */
}


/* Number of iterations to run A and B for. */
int nIter=5;

/* Stores state (registers and stack) for both threads */
ucontext_t A,B;

void runA(void *ptr) {
	int i,n;
	for (n=0;n<nIter;n++) {
		for (i=0;i<1000*1000;i++) {}
		printf("A (%d)\n",n);
		??
	}
}

void runB(void *ptr) {
	int i,n;
	for (n=0;n<nIter;n++) {
		for (i=0;i<1000*1000;i++) {}
		printf("      B (%d)\n",n);
		??
	}
}

int main() {
	makeThread(&A,runA,0);
	makeThread(&B,runB,0);
	setcontext(&A); /* start running thread A */
	return 0;
}
